#!/usr/bin/env node
//'use strict';

const gulp = require('gulp');
const less = require('gulp-less');
const Liftoff = require('liftoff');
const _ = require('lodash');
const path = require('path');
const concat = require('gulp-concat');
const rename = require('gulp-rename');
const babel = require('gulp-babel');
const uglify = require('gulp-uglify');
const browserify = require('browserify');
const vsstream = require('vinyl-source-stream');
const vsbuffer = require('vinyl-buffer')
const moment = require('moment');
const chalk = require('chalk');
const argv = require('minimist')(process.argv.slice(2));
const fs = require('fs');
const plumber = require('gulp-plumber');
const imagemin = require('gulp-imagemin');
const imageRecompress = require('imagemin-jpeg-recompress');
const pngquant = require('imagemin-pngquant');
const cache = require('gulp-cache');

const through = require('through-gulp');

const {fileSize} = require('./plugins/file');

var config;

const timeFormat = "YYYY-MM-DD hh:mm:ss";

var webtool = new Liftoff({
    name: 'webtool', // 命令名字
    processTitle: 'webtool', //同上即可
    moduleName: 'webtool', //同上即可
    configFiles: {
        'toolconf.json': {
            cwd: '.'
        }
    },
    extensions: {
        '.js': null
    }
});

const commands = [
    {command:"init", label:"初始化,创建一个配置文件"},
    {command:"listen", label:"开始监听配置的变化"}
]

function ctimestr(){
    return chalk.yellow(moment().format(timeFormat));
}

//启动命令
webtool.launch({
        cwd: argv.r || argv.root, //当前输入命令的目录
        configPath: argv.f || argv.file //读取当前命令目录下的配置文件路径
    }, function(env, enc) {

        //捕获命令行
    var command = argv["_"];
    if (argv.h){
        help();
    }
    if (command.length > 0)
    {
        switch (command[0]){
            case "init":
                init(env);
                break;

            case "listen":
                listen(env);
                break;

            default:
        }
    }


});

function init(env){
    var jsonPath = path.join(process.mainModule.filename, "../../toolconf.json");
    config = JSON.parse(fs.readFileSync(jsonPath.toLocaleString()));

    gulp.src(jsonPath.toLocaleString())
        .pipe(plumber())
        .pipe(gulp.dest('./'))
        .on("finish", function (){
            console.log('create config json file!');
        })

}
function listen(env){
    var config = require(path.join(env.cwd, 'toolconf.json'));
    var queue = config.convert;
    _.forEach(queue, function(obj){
        if(obj.source){
            const watchTarget = obj.watch || obj.source;
            switch (obj.type.toLowerCase()){
                case "css":
                    converCss(env, obj.source, obj.dest, obj.nano);
                    gulp.watch(watchTarget, function(e){
                        logEvent(e);
                        converCss(env, obj.source, obj.dest, obj.nano);
                    });
                    break;
                case "js":
                    converJs(env, obj.source, obj.dest);
                    gulp.watch(watchTarget, function(e){
                        logEvent(e);
                        converJs(env, obj.source, obj.dest);

                    });
                    break;
                case "browserify":
                    converBrowserifyJs(env, obj.source, obj.dest);
                    gulp.watch(watchTarget, function(e){
                        logEvent(e);
                        converBrowserifyJs(env, obj.source, obj.dest);
                    });
                    break;
                case "image":

                    if (obj.tinify){
                        compressImagesTinify(env, obj.source, obj.dest, obj.tinify);
                        gulp.watch(watchTarget, function(e){
                            logEvent(e);
                            compressImagesTinify(env, watchTarget, obj.dest, obj.tinify);
                        });
                    }
                    else{
                        compressImages(env, obj.source, obj.dest);
                        gulp.watch(watchTarget, function(e){

                            logEvent(e);
                            compressImages(env, watchTarget, obj.dest);
                        });
                    }
                    break;

            }
        }

    });
}

function logEvent(e){
    switch (e.type){
        case "added":
            break;
        case "changed":
            break;

        case "deleted":
            break;
    }
    console.log(`####${e.type} ${e.path} @${ctimestr()}####`)
}

function converCss(evn, source, dest, nano){

    var sourceCapacity = 0, destCapacity = 0;
    var headMsg = `<============conver process begin @${ctimestr()}==================\n`;
    var middMsg = '\t\t\t||\n\t\t\t||\n\t\t\t||\n\t\t\t||\n\t\t\t\\\/\n';

    var lessobj = less({
        strictMath:true
    });

    const cssnano = require('gulp-cssnano');
    const extname = path.extname(dest);
    const basename = path.basename(dest);
    var stream = gulp.src(source)
        .pipe(plumber())
        .on('data', function(file){
            var capacity = file.contents.length;
            sourceCapacity += file.contents.length;
            headMsg += `${file.path}(${fileSize(capacity)})\n`
        })
        .pipe(lessobj);

    if(extname.length > 0){
        stream = stream.pipe(concat(basename));
        destPath = path.dirname(dest);
    }
    else {
        destPath = dest
    }
    let nanoConfig = nano || {};
    stream = stream.pipe(cssnano(nanoConfig))
        .pipe(rename(function(path){
            path.basename += '.min';
        }))
        .pipe(gulp.dest(destPath))
        .on('data', function(file){
            var capacity = file.contents.length;
            destCapacity += capacity;
            middMsg += `${file.path}(${fileSize(capacity)})\n`

        })
        .on('finish', function(file){
            var endMsg = `==============conver process end @${ctimestr()}================>\n`
            console.log(headMsg);
            console.log(middMsg);
            msgConverResult(sourceCapacity, destCapacity);
            console.log(endMsg);
        });
    return stream;
}
function converJs(evn, source, dest){
    const extname = path.extname(dest);
    const basename = path.basename(dest);
    var stream;
    var sourceCapacity = 0, destCapacity = 0;
    var headMsg = `<============conver process begin @${ctimestr()}==================\n`;
    var middMsg = '\t\t\t||\n\t\t\t||\n\t\t\t||\n\t\t\t||\n\t\t\t\\\/\n';

    stream = gulp.src(source)
        .pipe(plumber())
        .on('data', function(file){
            var capacity = file.contents.length;
            sourceCapacity += file.contents.length;
            headMsg += `${file.path}(${fileSize(capacity)})\n`
        });
    stream = stream.pipe(plumber()).pipe(uglify());
    var destPath;

    if(extname.length > 0){
        stream = stream.pipe(concat(basename));
        destPath = path.dirname(dest);
    }else{
        destPath = dest
    }
    stream = stream.pipe(gulp.dest(destPath))
        .on('data', function(file){
            var capacity = file.contents.length;
            destCapacity += capacity;
            middMsg += `${file.path}(${fileSize(capacity)})\n`

        })
        .on('finish', function(){

            var endMsg = `==============conver process end @${ctimestr()}================>\n`
            console.log(headMsg);
            console.log(middMsg);
            msgConverResult(sourceCapacity, destCapacity);
            console.log(endMsg);
        });


    return stream;
}

function converBrowserifyJs(env, source, dest){
    const extname = path.extname(dest);
    const destfolder = path.dirname(dest);
    const babelify = require('babelify');
    if (extname.length == 0) return;

    var pluginObjectassign = require('babel-plugin-transform-object-assign');
    //var externalhelpers = require('babel-plugin-external-helpers');
    var es2015class = require('babel-plugin-transform-es2015-classes');
    var es2015modules = require('babel-plugin-transform-es2015-modules-commonjs');
    var preset2015 = require('babel-preset-es2015');
    var presetstage2 = require('babel-preset-stage-2');
    var presetreact = require('babel-preset-react');

    var sourceCapacity = 0, destCapacity = 0;
    var headMsg = `<============conver process begin @${ctimestr()}==================\n`;
    var middMsg = '\t\t\t||\n\t\t\t||\n\t\t\t||\n\t\t\t||\n\t\t\t\\\/\n';

    var b = browserify({
        entries: source
    });
    return b.transform(babelify, {  //此处babel的各配置项格式与.babelrc文件相同
            presets: [
                preset2015,  //转换es6代码
                presetstage2,  //指定转换es7代码的语法提案阶段
                presetreact  //转换React的jsx
            ],
            plugins: [
                pluginObjectassign,
                //externalhelpers,  //将es6代码转换后使用的公用函数单独抽出来保存为babelHelpers
                [es2015class, { "loose": false }],  //转换es6 class插件
                [es2015modules, { "loose": false }]  //转换es6 module插件
            ]
        })
        .bundle().on('error', function (err){
            console.log(err);
            this.emit('end');
        })
        .pipe(vsstream(path.basename(dest)))
        .pipe(vsbuffer())
        .on('data', function(file){
            var capacity = file.contents.length;
            sourceCapacity += file.contents.length;
            headMsg += `combine to visula file ${file.path}(${fileSize(capacity)})\n`
        })
        .pipe(uglify())
        .pipe(gulp.dest(destfolder))
        .on('data', function(file){
            var capacity = file.contents.length;
            destCapacity += capacity;
            middMsg += `${file.path}(${fileSize(capacity)})\n`

        })
        .on('finish', function(file){
            var endMsg = `==============conver process end @${ctimestr()}================>\n`
            console.log(headMsg);
            console.log(middMsg);
            msgConverResult(sourceCapacity, destCapacity);
            console.log(endMsg);
        });
}
function compressImages(evn, source, dest){

    var sourceCapacity = 0, destCapacity = 0;
    var headMsg = `<============conver process begin @${ctimestr()}==================\n`;
    var middMsg = '\t\t\t||\n\t\t\t||\n\t\t\t||\n\t\t\t||\n\t\t\t\\\/\n';
    gulp.src(source)
        .pipe(plumber())
        .on('data', function(file){
            var capacity = file.contents.length;
            sourceCapacity += file.contents.length;
            headMsg += `${file.path}(${fileSize(capacity)})\n`
        })
        .pipe(cache(imagemin({
            optimizationLevel: 3, //类型：Number  默认：3  取值范围：0-7（优化等级）
            progressive: true, //类型：Boolean 默认：false 无损压缩jpg图片
            interlaced: true, //类型：Boolean 默认：false 隔行扫描gif进行渲染
            multipass: true, //类型：Boolean 默认：false 多次优化svg直到完全优化
            use:[
                imageRecompress({
                    accurate: false,//高精度模式
                    quality: "low",//图像质量:low, medium, high and veryhigh;
                    method: "smallfry",//网格优化:mpe, ssim, ms-ssim and smallfry;
                    min: 60,//最低质量
                    strip:true,
                    target:0.1,
                    loops: 6,//循环尝试次数, 默认为6;
                    progressive: true,//基线优化
                    subsample: "default"//子采样:default, disable;
                    }
                ),
                pngquant({
                    optimizationLevel: 4
                })
            ]
        })))
        .pipe(gulp.dest(dest))
        .on('data', function(file){
            var capacity = file.contents.length;
            destCapacity += capacity;
            middMsg += `${file.path}(${fileSize(capacity)})\n`

        })
        .on('finish', function(){
            var endMsg = `==============conver process end @${ctimestr()}================>\n`
            console.log(headMsg);
            console.log(middMsg);
            msgConverResult(sourceCapacity, destCapacity);
            console.log(endMsg);
        });

}
function compressImagesTinify(evn, source, dest, config) {

    var sourceCapacity = 0, destCapacity = 0;
    var headMsg = `<============conver process begin @${ctimestr()}==================\n`;
    var middMsg = '\t\t\t||\n\t\t\t||\n\t\t\t||\n\t\t\t||\n\t\t\t\\\/\n';
    var tinify = require('./plugins/tinypng');
    if (!config.apiKey)
    {
        console.log('tinify need a api to compress images');
        return;
    }
    var keys = typeof config.apiKey == "string" ? [config.apiKey] : config.apiKey;

    gulp.src(source)
        .pipe(plumber())
        .on('data', function(file){
            var capacity = file.contents.length;
            sourceCapacity += file.contents.length;
            headMsg += `${file.path}(${fileSize(capacity)})\n`
        })
        .pipe(cache(tinify({
            keys:keys
        })))
        .pipe(gulp.dest(dest))
        .on('data', function(file){
            var capacity = file.contents.length;
            destCapacity += capacity;
            middMsg += `${file.path}(${fileSize(capacity)})\n`

        })
        .on('finish', function(){
            var endMsg = `==============conver process end @${ctimestr()}================>\n`
            console.log(headMsg);
            console.log(middMsg);
            msgConverResult(sourceCapacity, destCapacity);
            console.log(endMsg);
        });
}

function help(){
    console.log("webtool [<param>] -h");
    console.log("-h ——呼叫帮助");
    _.forEach(commands, function (obj){
        console.log(`<param>:${obj.command} -- ${obj.label}`);
    });
}
function msgConverResult(orginalSize, newSize){
    var dif = (orginalSize - newSize);
    //console.log(`${orginalSize}, ${newSize}`)
    var percent = Math.round(dif * 100 / orginalSize * 10) / 10;
    console.log(`save ${percent}% ====> ${fileSize(dif)}`);
}

//function pipeConsole(msg){
//    var stream = through(function(file, encoding,callback){
//        this.push(file);
//        callback();
//    });
//    console.log(msg);
//    return stream;
//}
